
#define TRACE

#ifdef TRACE
#include <math.h>
#endif

#include "may.h"

#define PIX_HIG 512
#define PIX_WID 512
#define PIX_LINE_LEN 256
#define PIX_PIECES  2	 /* essentially PIX_WID/PIX_LINE_LEN */


/* -------------------------------------------------------------- */

void main (argc, argv) 
int argc;
char **argv; 
{
   /*
    * This program implements a very crude algorithm for distributing and 
    * load balancing (and even fault tolerating!) the work neccessary to 
    * compute the the value of pixels in a pixel array. 
    * The pixel array is subdivided into horizontal strips of size 
    * 1 by PIX_LINE_LEN, which are then shipped off in arbitrary order to
    * other processors running MAY.
    *
    * Note that some extra code is interpersed to monitor statistics.
    *
    * Note that this program assumes that you are aquainted with the
    * Mandelbrot Set algorithm. The algorithm itself is pretty darn
    * simple, but if you don't understand it, some of the stuff below
    * might be cryptic.
    */
    static float	data [PIX_LINE_LEN];		/* my data array */
    static char		byte_data [PIX_LINE_LEN];	/* my data array */
    unsigned int	data_width, data_height;/* data array dimensions */
    float		zmax;
    unsigned int 	pix_width, pix_height; 	/* pixmap size */
    unsigned int	i_row, iarr, icol;

    int i;

    struct	mayProcId	*myself; 	/* my address */
    struct	mayProcId	*rem_proc [30];	/* remote process addresses */
    struct	mayMsghead		header;	/* header of returned message */
    int		messlen;		/* length of returned message */
    int 	retval;		/* mayRecv return code */

    /*
     * Note the data strucutre below is used both for storing the status
     * of the calculation and as the format of the message that is sent 
     * to the remote processes.
     *
     * Note, interestingly, that it has padding bits (16 bits worth) 
     * embedded in the middle. This was needed because this program
     * has been running in a heterogeneous environment of machines 
     * from a number of different vendors.  Different CPU architectures
     * mean that some machines have restrictions on the addresses that
     * things like words (long ints, floats, doubles, ...) can begin
     * on.  (Notably for the MIPS R3000 RISC chip).  In other words,
     * some compilers will stick those padding bits in for you,
     * whether you want them or not. 
     *
     */
    struct	initial_data {		/* message format and status return */
	unsigned int ix;	/* x start point of pixel line, in pixels */
	unsigned int iy;	/* y start point of pixel line, in pixels */
	unsigned int ilen;	/* length of one pixel line */
	unsigned int pad_bits;	/* padding bytes for sgi machines */
	double fx;	/* starting point, real part */
	double fy;	/* starting point, imaginary part */
	double delta;	/* grid increment, on the complex plane */
	unsigned int itermax;	/* maximum number of iterations to perform */
	int not_done;	/* flag indicating that this piece is done */
	} status [PIX_PIECES*PIX_HIG];


    char 	machine_name [30][80];	/* array containing machine names */
    char	file_name [30][80];	/* array containing file paths */
    long	queue_length [30];	/* queue length of each machine */
    short	machine_is_free [30];	/* array of free machines */
    long	low_water_mark;		/* almost empty marker */
    long	high_water_mark;	/* full marker */
    unsigned int   batch;		/* job batching counter */
    unsigned int   batch_block;    /* num of messages sent to each machine */
    unsigned int   num_machines;	/* number of available processors */
    unsigned int   num_free_machines;	/* number of free processors */
    unsigned int   curr_machine;	/* load balancing pointer */
    unsigned int   timeout_period;	/* how long to wait before timing out*/
    unsigned int   wait_for_work;	/* how long to wait before timing out*/
    unsigned int   not_timed_out;	/* receive fault flag */
    unsigned int   not_done;		/* finished flag */
    unsigned int   icnt;
    unsigned int   i_last_looked_at;

#ifdef TRACE
    /* variables used only for performance monitoring and trace */
    unsigned int   num_done;		/* number of completed tasks */
    unsigned int   num_since_last_timeout; /* num received since last timeout*/
    unsigned int   num_of_timeouts;	 /* number of timeouts to completeion */
    unsigned int   num_duplicated; /* num of tasks mistakenly duplicated */
    unsigned int   percent_dups;	/* percent of duplicated tasks */
    unsigned int   num_sent;		/* number of sent tasks */
    unsigned int   whos_working[30]; 	/* tally of machines that did work */
    unsigned int   name_xlate[30];	/* x-refernce machine to machine name */
    unsigned int   avg_received_per_batch; /* avg answers back per timeout */
    int   dribble_start;	/* num of timeouts before dribble */
    unsigned int   num_done_pre_dribble; /* num jobs completed before dribble*/
    unsigned int   num_dup_pre_dribble; /* num duplicates before dribble*/
    long	elapsed_time;		/* elapsed time */
    int		seconds, minutes, hours; /* elapsed time */
    unsigned int   unique_done;		/* amount of unique work done */
    unsigned int   sum_queue_len[30];	/* sum of queue lengths */
    unsigned long  sum_square_len[30];	/* sum of squares of queue lengths */
    unsigned int   num_queue_reports[30]; /* num of times Q length reported */
#endif


    /* 
     * Mandelbrot set stuff -- coordinates on the complex plane,
     * real and imaginary parts, and so on 
     */
    double	re_center,   im_center;	/* center of data array */
    double	width;			/* width of data aray */
    double	delta;			/* distance between adjacent pixels */
    double	re_start, im_start;	/* upper left hand corner of array */
    unsigned int itermax;		/* max iteration count */

    /*
     * get the seed data for the mandelbrot strange repellor.
     */
    get_mandelbrot_stuff ( &re_center, &im_center, &width, &itermax);
    delta = width / (double) PIX_WID;
    re_start = re_center - width / 2.0;
    im_start = im_center + width * ((double) PIX_HIG) / ((double) 2*PIX_WID);
 
    /* ------------------------------------------------
     * Set some very important and relevent parameters.
     * These affect system response and thouroughput.
     * See below for explanation.
     */
    low_water_mark = 3;
    high_water_mark = 60;
    timeout_period = 1;	  /* seconds to wait before timing out */
    /* batch_block is the number of messages sent to each machine */
    batch_block = high_water_mark - low_water_mark;
    /* ------------------------------------------------ */

    /*
     * Read in the machine names and the file path names
     */
    get_machines_and_file_paths (machine_name, file_name);

    /* 
     * Initialize status array with work to be done 
     * For explanation of data fields, see above
     */
    for (i_row=0; i_row<PIX_HIG; i_row++) {
       for (icol=0; icol<PIX_PIECES; icol++) {
          iarr = i_row*PIX_PIECES + icol;
          status [iarr] .iy = i_row;
          status [iarr] .ix = icol*PIX_LINE_LEN;
          status [iarr] .ilen = PIX_LINE_LEN;
          status [iarr] .fx = re_start + delta * (double) (icol * PIX_LINE_LEN);
          status [iarr] .fy = im_start;
          status [iarr] .delta = delta;
          status [iarr] .itermax = itermax;
          status [iarr] .not_done = TRUE;
       }
       im_start -= delta;
    }

    /* create the window in which we will show results of work */
    initialize_RGB_window (argc, argv, PIX_WID, PIX_HIG);

#ifdef TRACE
    /* start measuring elapsed time after the window has been placed */
    elapsed_time = time (NULL);

    printf (" Master completed initializing \n");
#endif

    /* intialize my self for MAY */
    mayInitTcp ();

#ifdef TRACE
    printf (" Master completed MAY initialization\n\n");
#endif

    /* 
     * Start up the remote processes.
     * num_machines will count how many processes actually started up.
     */
    num_machines=0;
    for (i=0; machine_name[i][0] != NULL; i++) { 
       rem_proc [num_machines] = mayCreate (machine_name[i], 
                                            file_name[i],
                                            MAY_CREATE_TCP);
       if (rem_proc [num_machines] == NULL ) {
          printf (" Machine %s failed to respond \n", machine_name [i]);
       } else {
#ifdef TRACE
          whos_working [num_machines] = 0;
          num_queue_reports [num_machines] = 0;
          sum_queue_len [num_machines] = 0;
          sum_square_len [num_machines] = 0;
          name_xlate [num_machines] = i;
#endif
          queue_length [num_machines] = 0;
          machine_is_free [num_machines] = TRUE;
          num_machines++;
       }
    }
    num_free_machines = num_machines;
    curr_machine = 0;

    if (num_machines == 0) {
       printf ("Error: not able to launch ANY remote precesses \n");
       exit (EXIT_FAILURE);
    }

#ifdef TRACE
    printf ("\n\n");
    for (i=0; i<num_machines; i++) {
       printf (" Processes now running on %s \n", machine_name [name_xlate[i]]);
    }
    printf (" Master: done creating %d remote procs \n", num_machines);
    printf ("\n\n");

    /* start the data transmission and collection loop */
    num_done = 0;
    num_duplicated = 0;
    unique_done= 0;
    num_since_last_timeout = 0;
    num_of_timeouts = 0;
    dribble_start = -1;
#endif

    /* start the data transmission and collection loop */
    wait_for_work = timeout_period;
    not_done = TRUE;
    icnt = 0;
    while (TRUE) {

       /*
        * Send out a big glob of data, in round-robin fashion, to all
        * participating machines.  Because there is no pacing involved
        * here, this will absolutely FLOOD the network (and your token
        * ring / Ethernet device drivers) with messages.  Frankly
        * speaking, the device driver on your machine will choke on
        * this, and you can expect that a large number of the messages
        * sent out will die and dissappear on your very own machine.
(int the UDP version -- but this is now the TCP version
        * Some smaller fraction will die elsewhere on the network,
        * typically on gateway machines and wherever else there are 
        * bottlenecks.
        *
        * The size of the glob of data  sent out is given by the
        * "batch_block" variable.
        */

/* send out a small amount of work to each machine, and don't repeat the
work request till we've come around at least once */

       i_last_looked_at = icnt;
       for (i=0; i < batch_block*num_free_machines; ) {

          /* skip this machine if its queue is full */
          while (machine_is_free [curr_machine] == FALSE) {
             curr_machine ++;
             curr_machine %= num_machines;
          }

          if (status[icnt].not_done != FALSE) {
             i++;
             maySend (rem_proc [curr_machine], 37+curr_machine, (short) icnt, 
                   (char *) &status[icnt], sizeof (struct initial_data) );
             curr_machine ++; 
             curr_machine %= num_machines;
          }
          icnt ++;
          icnt %= PIX_PIECES*PIX_HIG;
          if ( icnt == i_last_looked_at ) {
              /* if icnt equals last-looked-at, we've come around once.
               * if i is zero, means we didn't find anything that wasn't
               * done. So we're outa here.
               */
              if ( i == 0 ) not_done = FALSE;
#ifdef TRACE
              /* keep a fresh set of statistics for the tail end of work */
              if (dribble_start < 0) {
                  dribble_start = num_of_timeouts;
                  num_done_pre_dribble = num_done;
                  num_dup_pre_dribble = num_duplicated;
               }
#endif
              break;
           }
       }
       if (!not_done) break;


       /* 
        * Ok, now receive the completed work.
        * The plan of attack here is to read and read and read, until
        * the read queue is empty (until the messages stop arriving).
        * After a period of 'timeout_period', (typically one second),
        * we stop reading, and go back and dump another batch of 
        * messages out * onto the network.  We keep doing this until 
        * eventually every last pixel of the picture has been computed
        * and received.
        */
       header.msgptr = (char *) &data[0];

       not_timed_out = TRUE;
       while (not_timed_out == TRUE) {
          retval = mayReceive (NULL, NULL, NULL, 
                               wait_for_work, &header, &messlen);
          if (retval == MAY_NO_OPEN_SOCKS) {
             printf ("Error: all remote processes died before completeion of tasks \n");
             exit (EXIT_FAILURE);
          }

          if (retval == MAY_TIMED_OUT) {
             not_timed_out = FALSE;
#ifdef TRACE
             /* count the number of times through the loop */
             num_of_timeouts ++;
             printf ("Master: %d th timeout", num_of_timeouts);
             printf (" num since last timeout= %d",
                                       num_since_last_timeout );
             printf (" total= %d unique= %d \n", num_done, unique_done);
             num_since_last_timeout = 0;
#endif
          }
          else {
             if (messlen != sizeof (data)) {
                printf (" message length error, got %d shoulda got %d \n",
			messlen, sizeof (data) );
             }

             /* 
              * Mark this data segment as having been processed 
              */
             iarr = (unsigned int) header.subtype;
#ifdef TRACE
             /* If its already been processed, count it as a duplicate. */
             if (status[iarr].not_done == FALSE) {
                num_duplicated ++;
             } else {
                unique_done ++;
             }
#endif
             /* mark this piece of work as having been completed */
             status[iarr].not_done = FALSE;

             /*
              * disable machines which have plenty of work queued up,
              * enable machines that almost done with thier stuff
              */
             i = header.msgtype - 37;
             if (header.queue_len < low_water_mark) {
                 if (machine_is_free [i] == FALSE) {
                     num_free_machines ++;
                     machine_is_free [i] = TRUE;
                     wait_for_work = timeout_period;
                 }
             }
             if (header.queue_len > high_water_mark) {
                 if (machine_is_free [i] == TRUE) {
                     num_free_machines --;
                     machine_is_free [i] = FALSE;
                 }
                 if (num_free_machines == 0) wait_for_work = 1800;
             }

#ifdef TRACE
             /* 
              * Count the amount of work done,
              * Keep stat on which machine is doing what work.
              */
             num_done++;
             num_since_last_timeout++;
             i = header.msgtype - 37;
             whos_working [i] ++;
             num_queue_reports [i] ++;
             sum_queue_len [i] += header.queue_len;
             sum_square_len [i] += header.queue_len * header.queue_len;

   printf (" %s has a queue length of %d \n", 
		machine_name[name_xlate[header.msgtype-37]], 
		header.queue_len);        

#endif

             /* 
              * Write data to window, so that we can watch this live.
              * --- This is fairly CPU expensive, and not a terribly
              * good piece of code, but it is fun to watch.
              */
             for (i=0; i<PIX_LINE_LEN; i++) 
                 byte_data[i] = (char) (255.0* data[i]);
             (void) add_block_to_window (byte_data, PIX_LINE_LEN, 1,
			status[iarr].ix, status[iarr].iy);
             /* handle exposure events */
             (void) check_events();

          }
       }
    }
 
    /* enter event loop, handle window exposure events */
    event_loop();

    /* if we got to here, we are done */
    /* tell the child procs to stop */

    for (curr_machine = 0; curr_machine < num_machines; curr_machine++) {
       maySend (rem_proc [curr_machine], 999, 0, (char *) NULL, 0);
    }

#ifdef TRACE
    printf ("\n\n\n");

    elapsed_time = time(NULL) - elapsed_time;
    seconds = elapsed_time %60;
    elapsed_time /= 60;
    minutes = elapsed_time %60;
    elapsed_time /= 60;
    hours = elapsed_time;
    printf (" elapsed time = %d h %d m %d s \n", hours, minutes, seconds);

    printf (" total number of pieces needed = %d \n", PIX_PIECES*PIX_HIG);
    printf (" total number of pieces done = %d \n", num_done);

    percent_dups = (100 * num_duplicated) / (num_done - num_duplicated);
    printf (" percent duplicated jobs %d \n", percent_dups);

    printf (" number of messages sent per machine per batch = %d \n",
                               batch_block);
    printf (" total number of messages sent per batch = %d \n",
                               batch_block * num_machines);
    avg_received_per_batch = num_done / num_of_timeouts;
    printf (" average number of messages received per batch = %d \n",
                               avg_received_per_batch);
    printf (" total number of batches sent = %d \n", num_of_timeouts);

    printf ("\n");
    printf (" number of batches sent before petering out at end = %d \n",
                               dribble_start);
    if (dribble_start > 0) {
        printf (" avg num of msgs recvd per batch (steady state mode) = %d \n",
          num_done_pre_dribble / dribble_start);

        percent_dups = (100 * num_dup_pre_dribble) / 
                (num_done_pre_dribble - num_dup_pre_dribble);
        printf(" percent duplicated jobs before dribble %d \n", percent_dups);
    }

    printf (" number of batches dribbled back at the end = %d \n", 
                               num_of_timeouts - dribble_start);
    printf (" avg num of msgs recvd per batch (final dribble) = %d \n",
      (num_done - num_done_pre_dribble) / (num_of_timeouts - dribble_start));

    percent_dups = (100 * num_duplicated - num_dup_pre_dribble) / 
                (num_done - num_done_pre_dribble - 
                     (num_duplicated - num_dup_pre_dribble));
    printf(" percent duplicated jobs in dribble %d \n", percent_dups);

    printf ("\n");
    for (i = 0; i < num_machines; i++) {
       printf (" %s did %d pieces, or %d percent of the jobs\n",
          machine_name[name_xlate[i]],
          whos_working[i],
          100 * whos_working[i] /num_done );
    }

    printf ("\n");
    for (i = 0; i < num_machines; i++) {
       float avg, rms;

       avg = ((float) sum_queue_len[i]) / ((float) num_queue_reports[i]);
       rms = ((float) sum_square_len[i]) / ((float) num_queue_reports[i]);
       rms -= avg*avg;
       rms = sqrt (rms);
       printf (" %s had average queue length of %f, rms change %f \n",
          machine_name[name_xlate[i]], avg, rms);
    }
#endif

}
