
#include "may.h"

#define PIX_HIG 512
#define PIX_WID 512
#define PIX_LINE_LEN 256
#define PIX_PIECES  2	 /* essentially PIX_WID/PIX_LINE_LEN */

/*
 * The data strucutre below is used 
 * as the format of the message that is sent 
 * to the remote processes.
 *
 * Note, interestingly, that it has padding bits (16 bits worth) 
 * embedded in the middle. This was needed because this program
 * has been running in a heterogeneous environment of machines 
 * from a number of different vendors.  Different CPU architectures
 * mean that some machines have restrictions on the addresses that
 * things like words (long ints, floats, doubles, ...) can begin
 * on.  (Notably for the MIPS R3000 RISC chip).  In other words,
 * some compilers will stick those padding bits in for you,
 * whether you want them or not. 
 *
 */
struct initial_data {		/* message format and datarr return */
    unsigned int ix;		/* x start point of pixel line, in pixels */
    unsigned int iy;		/* y start point of pixel line, in pixels */
    unsigned int ilen;		/* length of one pixel line */
    double fx;			/* starting point, real part */
    double fy;			/* starting point, imaginary part */
    double delta;		/* grid increment, on the complex plane */
    unsigned int itermax;	/* maximum number of iterations to perform */
    unsigned int roll_over;	/* when to roll over iteration count */
    } datarr [PIX_PIECES*PIX_HIG];


/* -------------------------------------------------------------- */

#ifdef ANSI_C
int plug_it_in (char *data)
#else
int plug_it_in (data)
char * data; /* returned data array */
#endif 
{
    static char	pixels [PIX_LINE_LEN];
    int i, ix, iy;
    float *fptr;

    /* strip out pixel X and Y location */
    ix = (int) ((struct initial_data *) data) -> ix;
    iy = (int) ((struct initial_data *) data) -> iy;

#ifdef TRACE
    printf (" called back for ix = %d iy = %d", ix, iy);
    printf (" ilen = %d\n", (int) ((struct initial_data *) data) -> ilen);
    printf (" fx = %f fy = %f\n", 
        (double) ((struct initial_data *) data) -> fx,
        (double) ((struct initial_data *) data) -> fy);
    printf (" delta = %f itermax = %d \n", 
        (double) ((struct initial_data *) data) -> delta,
        (int) ((struct initial_data *) data) -> itermax);
#endif

    /* 
     * Write data to window, so that we can watch this live.
     * --- This is fairly CPU expensive, and not really good 
     * coding practice, but it is fun to watch.
     */
    fptr = (float *) (data + sizeof (struct initial_data));
    for (i=0; i<PIX_LINE_LEN; i++) pixels [i] = (char) (255.0* (*(fptr+i)));

    (void) add_block_to_window (pixels, PIX_LINE_LEN, 1, ix, iy);

    /* handle exposure events */
    (void) check_events();
}

/* -------------------------------------------------------------- */

void main (argc, argv) 
int argc;
char **argv; 
{
   /*
    * This program implements a very crude algorithm for distributing and 
    * load balancing (and even fault tolerating!) the work neccessary to 
    * compute the the value of pixels in a pixel array. 
    * The pixel array is subdivided into horizontal strips of size 
    * 1 by PIX_LINE_LEN, which are then shipped off in arbitrary order to
    * other processors running MAY.
    *
    * Note that some extra code is interpersed to monitor statistics.
    *
    * Note that this program assumes that you are aquainted with the
    * Mandelbrot Set algorithm. The algorithm itself is pretty darn
    * simple, but if you don't understand it, some of the stuff below
    * might be cryptic.
    */

    struct	mayProcId	*myself; 	/* my address */

    char 	**machine_name;	/* array containing machine names */
    char	**file_name;	/* array containing file paths */
    int   num;	/* number of available processors */

    /* 
     * Mandelbrot set stuff -- coordinates on the complex plane,
     * real and imaginary parts, and so on 
     */
    double	re_center,   im_center;	/* center of data array */
    double	width;			/* width of data aray */
    double	delta;			/* distance between adjacent pixels */
    double	re_start, im_start;	/* upper left hand corner of array */
    unsigned int itermax;		/* max iteration count */
    unsigned int roll_over;             /* when to roll the float over */

    int	i, irow, icol, iarr;

    /*
     * get the seed data for the mandelbrot strange repellor.
     */
    get_mandelbrot_stuff ( &re_center, &im_center, 
                           &width, &itermax, &roll_over);
    delta = width / (double) PIX_WID;
    re_start = re_center - width / 2.0;
    im_start = im_center + width * ((double) PIX_HIG) / ((double) 2*PIX_WID);
#ifdef TRACE
    printf (" (re, im) center = (%f %f) \n", re_center, im_center);
    printf (" (re, im) start = (%f %f) \n", re_start, im_start);
    printf (" width = %f delta = %f \n", width, delta);
#endif
 
    /* intialize my self for MAY */
#ifdef UDP
    myself = mayInitUdp ();
#else
    mayInitTcp ();
#endif

    /*
     * Read in the machine names and the file path names
     */
    num = get_machines_and_file_paths (&machine_name, &file_name);

    /* 
     * Start up the remote processes.
     */
#ifdef UDP
    mayCreateFarm (num, machine_name, file_name, MAY_CREATE_UDP);
#else
    mayCreateFarm (num, machine_name, file_name, MAY_CREATE_TCP);
#endif

    /* 
     * Initialize datarr array with work to be done 
     * For explanation of data fields, see above
     */
    for (irow=0; irow<PIX_HIG; irow++) {
       for (icol=0; icol<PIX_PIECES; icol++) {
          iarr = irow*PIX_PIECES + icol;
          datarr [iarr] .iy = irow;
          datarr [iarr] .ix = icol*PIX_LINE_LEN;
          datarr [iarr] .ilen = PIX_LINE_LEN;
          datarr [iarr] .fx = re_start + delta * (double) (icol * PIX_LINE_LEN);
          datarr [iarr] .fy = im_start;
          datarr [iarr] .delta = delta;
          datarr [iarr] .itermax = itermax;
          datarr [iarr] .roll_over = roll_over;
       }
       im_start -= delta;
    }

    /* create the window in which we will show results of work */
    initialize_RGB_window (argc, argv, PIX_WID, PIX_HIG);

#ifdef TRACE
    printf (" Master completed MAY initialization\n\n");
#endif

    mayProcFarm (datarr, PIX_PIECES*PIX_HIG, sizeof (struct initial_data),
       sizeof (struct initial_data),
       plug_it_in);
 
    /* if we got to here, we are done */
    /* tell the child procs to stop */
/*
    for (i = 0; i < num; i++) {
       maySend (rem_proc [i], 999, 0, (char *) NULL, 0);
       }
*/

    mayTerminate ();

    /* enter event loop, handle window exposure events */
    event_loop();

    }

/* -------------------------------------------------------------- */
