/* =================================================================
 *
 * Copyright (C) 1989, 1990, 1993 Linas Vepstas
 *
 *                   All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its 
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies 
 * and that both the copyright notice and this permission notice 
 * appear in supporting documentation.
 *
 * Permission to redistribute this software and its documentation
 * is hereby granted, provided that such redistribution is not 
 * performed for monetary profit or gain.
 *
 * Redistribution for profit without express written consent is
 * prohibited.  Contact the author if you wish to use this software 
 * and/or its documentation in a commercial product.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND.  
 * THE AUTHOR DISCLAIMS ALL EXPRESS AND IMPLIED WARRANTIES, INCLUDING 
 * BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL THE AUTHOR BE HELD LIABLE FOR ANY SPECIAL, DIRECT, 
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING 
 * IN LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Bug reports or fixes may be sent to the author, who may or may not 
 * act upon them as he desires.
 * 
 * The author's last known address is:
 * 
 *                       Linas Vepstas
 *                     1518 Enfield Road
 *                  Austin, Texas 78703-3424
 *                      1-(512)-499-8246
 *                    linas@austin.ibm.com
 *
 * =================================================================
 */


#include "may.h"
#include "maymsg.h"

int may_internal_flag;	/* flag to use NULL message type for internal reads */
long	may_msg_list_length = 0;	/* receive queue length */

#ifdef ANSI_C
int mayMatchMsg (void *byte_stream,
                 struct mayProcId *sproc_ptr,
                 long msg_type,
                 short sub_type);
#else
int mayMatchMsg ();
#endif

#define UDP_BUFF_LEN 	2048
char udp_recv_buff [UDP_BUFF_LEN];

/* ========================================================= */
/*  Portability thing for select call */

#define MNFDBITS	BITS(int)
#define MFD_SET(n, p)    ((p)->fdsmask[(n)/MNFDBITS] |= (1 << ((n) % MNFDBITS)))
#define MFD_CLR(n, p)    ((p)->fdsmask[(n)/MNFDBITS] &= ~(1 << ((n) % MNFDBITS)))
#define MFD_ISSET(n, p)  ((p)->fdsmask[(n)/MNFDBITS] & (1 << ((n) % MNFDBITS)))
#define MFD_ZERO(p)      memset((char *)(p), 0, sizeof(*(p)))


/* ========================================================= */
/* maySelect 
 * Performs select on open file descriptors
 *
 * This is implemented as a macro because I didn't want to have
 * the overhead of a subroutine call
 */

#define maySelect(timeout)					\
   {								\
      read_select_timeout.tv_sec = (int) timeout;		\
      read_select_timeout.tv_usec = 0;				\
      MFD_ZERO (&read_select_mask);				\
      select_return = 0;					\
      /* traverse the list of open sockets */			\
      os = open_socks;						\
      while (os) {						\
         MFD_SET (os->socket_id, &read_select_mask);		\
         os = os -> next;					\
      }								\
      select_return = select (NFDS(5*sizeof(int)*8), &read_select_mask, \
         NULL, NULL, &read_select_timeout);			\
      if (select_return == -1) {				\
         perror ("mayReceive error: select");			\
         return (MAY_GENERAL_ERROR);				\
      }								\
   }

/* ========================================================= */
/* mayCheckRecv 
 * Checks for closed connections
 *
 * This is implemented as a macro because I didn't want to have
 * the overhead of a subroutine call
 */

#define mayCheckRecv(rcv_len)					\
{								\
   if (rcv_len < 0) {						\
      perror ("mayCheckRecv error: recvfrom");			\
      return (MAY_GENERAL_ERROR);				\
   }								\
   else if (rcv_len == 0) {					\
      /* if the received length is zero, it means that the	\
       * socket at the other end shut down.  (This appears	\
       * to be standard UNIX way of indicating a closed		\
       * pipe and/or file).  Therefore, we may as well		\
       * close things up at this end too.			\
       */							\
      return (MAY_CLOSE_CONNECTION);				\
   }								\
}

/* ========================================================= */
/* mayCloseConnection
 * Removes a socket from our list of open socks
 *
 */

#define mayCloseConnection()					\
{								\
   struct mayProcId *os_next;					\
								\
   close (os->socket_id);					\
   if (os_previous != NULL) {					\
      os_previous->next = os->next;				\
   }								\
   if (os == open_socks) {					\
      open_socks = open_socks -> next;				\
   }								\
   os_next = os -> next;					\
   free (os->hostname);						\
   free (os->partial_msg);					\
   free (os);							\
   os = os_next;						\
   fprintf (stderr, "mayRecv: Warning: lost connection \n");	\
}


/* ========================================================= */
/* mayReceive 
 * Returns message to caller
 *
 * If sproc, msgtype and subtype are specified, it will return the first
 * message whose header matches these.  If any and/or all of sproc, msgtype,
 * subtype are NULL, a match is not made to these.
 * 
 * If the message is not received by timeout seconds, return with timeout
 * failure.  This is implemented with a select, which seems to work
 * very well, but is not quite portable unix ???
 * 
 * The message is returned in the area pointed to by mess.  The user must
 * allocate that area.  The size of the area allocated must be passed in
 * messlen.  Returned in messlen is the size of the message (in bytes), NOT
 * including the header (i.e. size of message body only).
 * 
 * The header should show the true senders address (NOT IMPLEMENTED)
 * 
 * psuedocode:
 * Empty out the udp and tcp system queues by reading indiscriminately.
 * Examine list of previously received messages.
 * 	Does type match?
 * 	Does subtype match?
 * 	Does sproc match?
 * 	Yes-- copy message into user space, dealloc list memory. Return.
 * 	No -- examine next message.
 * End of list?
 * Receive next message.
 *       Is there a message to be received? 
 *       Yes -- goto "empty out system queue"
 *       No -- continue
 *       Did we time out on the read?
 *       If timed out, return to user with timeout error.
 * 
 * 
 * Note that the linked list of recieved messages is kept in FIFO order.
 * Note that we will keep track of the length of this list; it can become a
 * potentially infinitely long list if the sender bombards us with
 * messages.  Keeping track of the queue length and sending that back to
 * the sender will help the sender implement a FIFO pacing algorithm.
 * 
 */

struct may_msg_list *may_msg_head_ptr = NULL;
struct may_msg_list *may_msg_tail_ptr = NULL;


#ifdef ANSI_C
int mayReceive (struct mayProcId *sproc_ptr,
                long 		msg_type,
                short	 	sub_type,
                unsigned int 	timeout,
                struct mayMsghead 	*mess,
                int 		*messlen)
#else
int mayReceive (sproc_ptr, msg_type, sub_type, timeout, mess, messlen)
		struct mayProcId *sproc_ptr;
		long 		msg_type;
		short	 	sub_type;
		unsigned int 	timeout;
		struct mayMsghead 	*mess;
		int 		*messlen;
#endif
{

   int retval;
   struct 	may_msg_list 	*search, *previous, *tmp;
   struct mayProcId *os;

#ifdef DEBUG
   if (open_socks == NULL) printf(" NO open sockets!!!\n");
#endif

   /* 
    * first we empty out the system queue of messages to be read
    * These will be chained onto our message list
    */
   retval = mayEmptyQueue (0);
   if (retval < 0) return (retval);


   while (TRUE) {
      /* Now that we've emptied the system queues, lets look at what we
       * got.  Hopefully we'll come up with something */
      /* previous will be used to rechain the list if match is found */
      previous = NULL;  
      for (search=may_msg_head_ptr; search!=NULL; search = search->next) {
   
         /* compare header to desired header */
         if (TRUE == mayMatchMsg (search->msg, sproc_ptr, msg_type, sub_type)) {
            /* We've got a match !! */
            mayBreakOutMsg (mess, messlen, search->msg);
            /* rechain */
            /* if the first message on the list */
            if (previous == NULL) {
               may_msg_head_ptr = may_msg_head_ptr -> next;
            }
            else { /* not the first on the list */
               previous -> next = search -> next;
            }
            /* if the last message on the list, back up one */
            if (search == may_msg_tail_ptr) {
               may_msg_tail_ptr = previous;
            } 
            free (search -> msg);  /* free message */
            free (search);    /* free node */
   
            /* update queue length counter */
            may_msg_list_length--;

            Tpr(("mayreceive found a message, exit with success !\n"));
            return (may_msg_list_length);  /* we've got it-- return with success! */
         }
         previous = search;
      }
   
      Tpr((" mayReceive: no matches found, try to time out in %d secs\n", 
            timeout ));
      /* 
       * If we got to here, it means that we didn't find the message in
       * our the queue. Thus, we will try real hard to receive it, until we
       * either do or we time out.
       *
       * Note also: if there are no open sockets, we should let user know
       * (Warning, not an error)
       */
      if (open_socks == NULL) {
         fprintf (stderr, "mayRecv: Warning: no open sockets! \n");
         return (MAY_NO_OPEN_SOCKS);
      }
      retval = mayEmptyQueue (timeout);
      if (retval ==  0) return (MAY_TIMED_OUT);
      if (retval < 0) return (retval);
   
      /* if we got to here, there must obviously be something to read 
       * so loop back & read it */
   }
}

/* ========================================================= */
/* mayReadMsgUdp
 * Read precisely one UDP message off of a socket.
 *
 * Read a message, check for message length errors, and copy 
 * over into our internal message queue.
 */

#ifdef ANSI_C
int mayReadMsgUdp (struct mayProcId *os, 
                   Void **recv_buff)
#else
int mayReadMsgUdp (os, recv_buff)
		struct mayProcId *os;
		Void **recv_buff;
#endif
{
   struct 	sockaddr_in spaddr;	/* address of sending process */
   int 		spaddrlen;		/* size of spaddr */
   int		rcv_len;
   int		mlen;

   /* DGRAM messages come in packets; 
    * header and message are in same packet */
   spaddrlen = SIZEOSOX;
   memset (&spaddr, 0, SIZEOSOX);
   rcv_len = recvfrom (os->socket_id, udp_recv_buff, 
      UDP_BUFF_LEN, 
      0, &spaddr, &spaddrlen);

   /*check for error conditions */
   mayCheckRecv(rcv_len);
   if (rcv_len < sizeof (struct general_msg_hdr)) {
      fprintf (stderr, 
      "mayReadMsgUdp error: message truncated to %d\n", rcv_len); 
      return (MAY_TRUNCATED_MSG);
   }

   mlen = ((struct general_msg_hdr *) udp_recv_buff) -> msglen;
   /* DO NOT change mesglen!!
    * important to have mlen just right when exiting */
   if (mlen != rcv_len) {
      if (mlen >= UDP_BUFF_LEN) {
         fprintf (stderr, 
            "mayReadMsgUdp error: message too long: %d\n", mlen);
      }
      fprintf (stderr, "mayReadMsgUdp error: message truncated, ");
      fprintf (stderr, " got %d should be %d\n", rcv_len, mlen); 
      return (MAY_TRUNCATED_MSG);
   }

   *recv_buff = (Void *) malloc (rcv_len);
   if (NULL == *recv_buff) {
      perror("mayReadMsgUdp: failure to malloc");
      return (MAY_NO_MEM);
   }

   memcpy (*recv_buff, udp_recv_buff, rcv_len);

   return (rcv_len);
}

/* ========================================================= */
/* mayReadMsgTcp
 * Read precisely one TCP message off of a socket.
 *
 * Read a message, check for message length errors, and copy 
 * over into our internal message queue.
 */

#ifdef ANSI_C
int mayReadMsgTcp (struct mayProcId *os, 
                   Void **recv_buff)
#else
int mayReadMsgTcp (os, recv_buff)
		struct mayProcId *os;
		Void **recv_buff;
#endif

{
   struct 	sockaddr_in spaddr;	/* address of sending process */
   int 		spaddrlen;
   struct general_msg_hdr  mh;
   int		rcv_len;
   int		mlen;
   char		*tmp_ptr;
   int		pre_len;	/* partial message length before read */
   int		new_len;	/* partial message length after read */


#define TCP_READ_DESIGN_A
#ifdef TCP_READ_DESIGN_A
   /* STREAM messages have no packet boundries;
    * so get the header first, see how long the message
    * is, and then try to get that. */
   spaddrlen = SIZEOSOX;
   memset (&spaddr, 0, SIZEOSOX);

   Tpr (("entering MayReadMsgTcp,unread bytes = %d \n",
         os -> unread_bytes_remaining ));

   if ((os->unread_bytes_remaining == 0) && !(os->is_header_read)) {
      os->bytes_read = 0;
      os->unread_bytes_remaining = sizeof (struct general_msg_hdr);
      os->partial_msg = (char *) malloc (sizeof (struct general_msg_hdr));
      if (os->partial_msg == NULL) {
         perror ("mayReadMsgTcp: unable to malloc");
         return (MAY_NO_MEM);
      }
   }
   rcv_len = recvfrom (os->socket_id, &(os->partial_msg)[os->bytes_read],
      os->unread_bytes_remaining, 0, &spaddr, &spaddrlen);

   if (rcv_len < 0) {
      perror ( "mayReadMsgTcp error" );
      return (MAY_GENERAL_ERROR);
   }


   Tpr (("mayReadMsgTcp: got %d bytes \n", rcv_len));

   /* if the received length is zero, it means that the socket at 
    * the other end shut down.  (This appears to be standard UNIX 
    * away of indicating a closed pipe and/or file).  Therefore, 
    * we may as well close things up at this end too.  */
   if (rcv_len == 0) {
      return (MAY_CLOSE_CONNECTION);
   }

   /* If we didn't receive as many bytes as we asked for, return. 
    * Later we'll come back and try again.  Remeber, this is a TCP 
    * connection, so we shouldn't be suffering from lost bytes or 
    * incomplete messages.  The worst that can happen is that the 
    * socket shut down before completeing transmission, in which
    * case we should reset. */
   os->bytes_read += rcv_len;
   os->unread_bytes_remaining -= rcv_len;
   if (os->unread_bytes_remaining > 0) {
      return (MAY_INCOMPLETE_READ);
   }

   /* if we got to here, we must have read all of the bytes.
    * Does this mean that we're done with the header, or 
    * do we have the whole message? */
   if (! (os->is_header_read)) {
      os -> is_header_read = TRUE;
      Tpr (("mayReadMsgTcp: Got the header \n"));

      /* OK, now, realloc for the proper message size */
      mlen = ((struct general_msg_hdr *) os->partial_msg) -> msglen;
      os->partial_msg = (char *) realloc (os->partial_msg, mlen);
      if (os->partial_msg == NULL) {
         perror ("mayReadMsgTcp: unable to malloc");
         return (MAY_NO_MEM);
      }
      os->unread_bytes_remaining = mlen - os->bytes_read;

      /* This is the lazy man's way of coding.  This will cause 
       * the code to loop back and try again.  If we weren't so 
       * lazy, we'd try to read again here, since we'd have a 
       * pretty good chance that there was more to be read.  To
       * be safe, we'd have to check with select again ... ugh.
       */
      return (MAY_INCOMPLETE_READ);
   }
   /* else, if we got to here, it means that all bytes have been
    * read, and we have got a whole message. */
   else {
      Tpr (("mayReadMsgTcp: Got the body \n"));
      mlen = os -> bytes_read;
      os -> bytes_read = 0;
      os -> is_header_read = FALSE;
      *recv_buff = os->partial_msg;
      os -> partial_msg = NULL;
      return  (mlen);
   }

#endif  /* TCP_READ_DESIGN_A */

}



/* ========================================================= */
/* mayEmptyQueue 
 * This routine empties out the system message buffer.
 * 
 * Note that both stream (TCP) and datagram (UDP) buffers are emptied. 
 * For UDP, this is probably a good idea: the UDP buffers have a tendency
 * to overflow, and operating systems tend to throw out UDP data when
 * overflows occur.  For TCP, this seems to be pretty wasteful, because we
 * know that the system won't loose the messages, AND, this is a second
 * message buffer when we really only need one (the system's queue).
 * 
 * However, the reson we do this is because we want to keep track of the
 * queue depth, so that we can inform the sender.  The sender can use this
 * information to its benefit (load balancing and FIFO pacing).
 * 
 * Returns:
 * number of messages read until timeout occurs.  If the number of messages is
 * zero,  then a true timeout has occured, since no messages have been read.
 *
 * Psuedocode:
 * Basically, it reads indiscriminantly from the file descriptor
 * socket.
 *
 * psuedocode:
 *      recieve; 
 * 	malloc memory; 
 *	copy message to memory; 
 *	append to list.
 * 	Loop back until select craps out.
 */

#ifdef ANSI_C
int mayEmptyQueue (int timeout)
#else
int mayEmptyQueue (timeout) 
int timeout;
#endif
{
   struct mayProcId *os, *os_previous;

   int		select_return;
   struct {	int fdsmask[5];		/* this structure needed by    */
      		} read_select_mask; 	/* the select command          */
   struct timeval read_select_timeout; /* select timeout */

   Void		*recv_buff;
   int		num_received;
   int		retval;			/* return code */

   maySelect (timeout);
   Tpr (("mayEmptyQueues: select says %d file desc's can be read\n", 
         select_return));

   num_received = 0;
   /* keep looping while there are messages to read */
   while (select_return != 0) {
      num_received ++;
      /* 
       * if we got to here after all that select call, there must be 
       * something to read 
       * traverse the list of open sockets
       */
      os = open_socks;
      os_previous = NULL;
      while (os) {
         if (!MFD_ISSET (os->socket_id, &read_select_mask)) {
            Tpr ((" mayEmptyQueues: there are zip messages on fd %d\n", 
               os->socket_id));
            os_previous = os;
            os = os->next;
            continue;
         }
         Tpr (( " mayEmptyQueues: select said there's stuff on fd %d address %d 0x%x %d \n",
            os->socket_id,
            os->address.sin_family, 
            os->address.sin_addr.s_addr, 
            os->address.sin_port));
         /* read one, and only one message */
         /* Use different types of recv's, depending on the protocol type */
         switch (os->proto) {
            case SOCK_DGRAM:
               retval = mayReadMsgUdp (os, &recv_buff);
               break;

            case SOCK_STREAM:
               retval = mayReadMsgTcp (os, &recv_buff);
               /* if an incomplete message was returned, go off and check up on
                * the other socks, and try again later. */
               if (retval == MAY_INCOMPLETE_READ) {
                  os_previous = os;
                  os = os -> next;
                  continue;
               }
               break;

            default:
               perror ("mayEmptyQueue error: unrecognized protocol type");
               return (MAY_GENERAL_ERROR);
         }

         /* if the socket died during read, 
          * clean up and move to next socket */
         if (retval == MAY_CLOSE_CONNECTION) {
            mayCloseConnection ();
            continue;
         }

         /* if error condition, return with error condition, 
          * else go on */
         if (retval < 0) return (retval);

/*
         Tpr (( " mayEmptyQueue got one from address %d 0x%x %d ", 
         spaddr.sin_family, spaddr.sin_addr.s_addr, spaddr.sin_port));
*/
         Tpr (( " mayEmptyQueue got one "));
         Tpr (( " internally identified as %d 0x%x %d ", 
         ((struct general_msg_hdr *) recv_buff)->sproc.address.sin_family,
         ((struct general_msg_hdr *) recv_buff)->sproc.address.sin_addr.s_addr, 
         ((struct general_msg_hdr *) recv_buff)->sproc.address.sin_port));
         Tpr (( " with mess type %d subtype %d \n", 
         ((struct general_msg_hdr *) recv_buff)->msgtype,
         ((struct general_msg_hdr *) recv_buff)->subtype));

         /* insert the correct socket fd */
         ((struct general_msg_hdr *)recv_buff)->sproc.socket_id = os->socket_id;

         /* chain it on to our store */
         if (!mayChainMsg (recv_buff)) {
            perror ("mayEmptyQueue error: mayChainMsg malloc");
            return (MAY_NO_MEM);
         }

         /* increment queue length counter */
         may_msg_list_length++;
         os_previous = os;
         os = os -> next;
      }

      /* Lets check again --- any messages to read ?? */
      maySelect(0);

      Tpr ((" mayEmptyQueues select says file desc %d can be read\n", 
             select_return));

   }
   Tpr (("mayEmptyQueues: num reads = %d msgs cached = %d \n", 
                                       num_received, may_msg_list_length));
   return (num_received);
}


/*---------------------- Experminetal Code --------------------------*/

/*    printf (" mayRec: fd %d len, %d, err, %d %d \n", may_udp_sockfd, 
      rcv_len, errno, EWOULDBLOCK); 
*/
      /* if there were no messages to be received, wait a timeout */
      /* period and try again */
 /*   if ((rcv_len < 0) && (errno = EWOULDBLOCK) && 
         (timed_out_on_read == 0)) {
         timed_out_on_read = 1;
         sleep (timeout);
         continue;
         }
      if ((rcv_len < 0) && (errno = EWOULDBLOCK) && 
         (timed_out_on_read == 1)) {
         return (MAY_TIMED_OUT);
         }
      timed_out_on_read = 0;
 */

/* ========================================================= */
/* mayBreakOutMsg 
 * Copy byte stream into user format
 *
 * The routine mayBreakOutMsg copies the byte stream into the user message
 * format.  It is assumed that memory has already been allocated for the 
 * copy destination.
 */

#ifdef ANSI_C
void mayBreakOutMsg (struct mayMsghead *usr_mess,
                     int *usr_messlen,
                     char *byte_stream)
#else
void mayBreakOutMsg (usr_mess, usr_messlen, byte_stream)
		struct mayMsghead *usr_mess;
		int *usr_messlen;
		char *byte_stream;
#endif
{
   struct general_msg_hdr *hdr_ptr;
   char *buff;
   int offset;

   /* Break out the header */
   hdr_ptr = (struct general_msg_hdr *) byte_stream;
   usr_mess -> msgtype = hdr_ptr -> msgtype;
   usr_mess -> subtype = hdr_ptr -> subtype;
   usr_mess -> msglen = hdr_ptr -> msglen - sizeof (struct general_msg_hdr);
   usr_mess -> queue_len = hdr_ptr -> queue_len;
   memcpy (&usr_mess->sproc, &hdr_ptr->sproc, SIZEOPRO);
   *usr_messlen = usr_mess -> msglen;

   /* copy the message into area pointed at */
   buff = (char *) byte_stream;
   offset = sizeof (struct general_msg_hdr);
   memcpy (usr_mess->msgptr, &buff [offset], usr_mess->msglen);

}

/* ========================================================= */
/* mayMatchMsg 
 * Scans for a matching message type
 * 
 * This routine compares the header pointed to by byte_stream with the 
 * passed sproc, message type and subtype.  It returns TRUE if a match
 * is found, else FALSE.
 * 
 * This routine has two modes of operation: "user mode" and "internal mode".
 * In user mode, it is assumed that argument of NULL for sproc, msgtype
 * or subtype means "don't care-- match to anything".  Note this implies
 * that its impossible to match to messages of msgtype zero (the msgtype
 * used internally by daemon).
 * In the internal mode, a check is made to msgtype zero, so that internal
 * messages can be detected.
 */

#ifdef ANSI_C
int mayMatchMsg (void *byte_stream,
                 struct mayProcId *sproc_ptr,
                 long msg_type,
                 short sub_type)
#else
int mayMatchMsg (byte_stream, sproc_ptr, msg_type, sub_type)
		/* void *byte_stream;   /* whats with this compiler and type void ??? */
		Void *byte_stream;
		struct mayProcId *sproc_ptr;
		long msg_type;
		short sub_type;
#endif
{
   struct general_msg_hdr *hdr_ptr;

   hdr_ptr = (struct general_msg_hdr *) byte_stream;

#ifdef DEBUG
   if (sproc_ptr != NULL) {
      printf ( " mayMatchMsg looking for msg from %d 0x%x %d ", 
      sproc_ptr -> address.sin_family,
      sproc_ptr -> address.sin_addr.s_addr,
      sproc_ptr -> address.sin_port);
   }
   else {
      printf (" mayMatchMsg looking for mess from anybody ");
   }
   printf (" of type %d subtype %d \n", msg_type, sub_type);

   printf ( " mayMatchMsg is looking at %d 0x%x %d ", 
      hdr_ptr -> sproc.address.sin_family,
      hdr_ptr -> sproc.address.sin_addr.s_addr, 
      hdr_ptr -> sproc.address.sin_port);
   printf ( " with mess type %d subtype %d \n", 
      hdr_ptr -> msgtype,
      hdr_ptr -> subtype);
#endif

   /* if we are using this routine "internally", we want to match to */
   /* message type zero.  */
   if (may_internal_flag != 0) { 
      if (DAEMONMSG != hdr_ptr->msgtype) return (FALSE); 
   }
   else {
      /* If msg_type not NULL, see if theres a match */
      if ((msg_type != 0) && (msg_type != hdr_ptr->msgtype)) return (FALSE);
   }

   /* If sub_type not NULL, see if theres a match */
   if ((sub_type != 0) && (sub_type != hdr_ptr->subtype)) return (FALSE);

   /* If sproc_ptr not NULL, see if theres a match */
   if ((sproc_ptr != NULL) && 
      (0 != memcmp (sproc_ptr->address, 
       &hdr_ptr->sproc.address, SIZEOPRO))) return (FALSE);

   Tpr (( " mayMatchMsg found a match to %d 0x%x %d ", 
      hdr_ptr -> sproc.address.sin_family,
      hdr_ptr -> sproc.address.sin_addr.s_addr, 
      hdr_ptr -> sproc.address.sin_port));
   Tpr (( " with mess type %d subtype %d \n", 
      hdr_ptr -> msgtype,
      hdr_ptr -> subtype));

   /* If we got to here, we survived all tests and must obviously match */
   return (TRUE);

}

/* ========================================================= */
/* mayChainMsg 
 * Add message to our internal list of messages
 *
 */

#ifdef ANSI_C
int mayChainMsg (void 	*buff)		/* receive buffer */
#else
int mayChainMsg (buff)
Void 	*buff;		/* receive buffer */
#endif

{
   struct 	may_msg_list 	*search, *previous, *tmp;

   /* chain it on to our store */
   tmp = (struct may_msg_list *) malloc (sizeof (struct may_msg_list));
   if (tmp == NULL) return (FALSE);

   tmp->msg = buff;
   /* chain the new message to the end of the current list */
   tmp -> next = NULL;
   if (may_msg_head_ptr == NULL){
      may_msg_head_ptr = tmp;
   }
   else {
      may_msg_tail_ptr -> next = tmp;
   } 
   may_msg_tail_ptr = tmp;

   /* sucess */
   return (TRUE);
} 

/* ====================== END OF FILE ========================== */
